/*****************************************************************************
 * File: crunch_WEB.c
 * 	(Extracted from
 *	   "$Id: crunchWEB.html,v 1.10 2001/05/19 11:36:12 ma Exp ma $")
 *
 * Abstracts:
 *  This program extracts source files from the HTML files.
 *  (The more informations:
 *     "http://sapporo.cool.ne.jp/maoya/crunchWEB/crunchWEB.html")
 *
 * Copyright (C) 1999 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 ****************************************************************************/

/**
 * Refer the Platform's Definitions
 **/
#include <unistd.h>	/* unistd.h*/
#include <stddef.h>	/* stddef.h*/
#include <stdio.h>	/* stdio.h*/
#include <errno.h>	/* errno.h*/
#include <limits.h>	/* limits.h*/
#include <ctype.h>	/* ctype.h*/
#include <string.h>	/* string.h*/
#include <sys/types.h>	/* sys/types.h */
#include <sys/stat.h>	/* sys/stat.h */
#include <sys/mman.h>	/* sys/mman.h */
#include <fcntl.h>	/* fcntl.h */

/**
 * DEBUG Level Switch ('0' means *NO-DEBUG*)
 **/
static int DEBUG = 0;

/**
 * This Program's Exit Codes
 **/
typedef enum EXIT_CODE_t {
	 EXIT_NORMAL = 0
	,EXIT_FAIL_FOPEN = 1
	,EXIT_FAIL_FPUTS
	,EXIT_FAIL_CLOSE_OUTPUT
	,EXIT_FAIL_CLOSE_OUTPUT_OUT_OF_PARSE_LOOP
	,EXIT_BROKEN_FSM_STATE_CONTROL
	,EXIT_FAIL_CLOSE_INPUT_STREAM
	,EXIT_FAIL_TO_MMAP_TMP_FILE
	,EXIT_FAIL_TO_UNMAP_TMP_FILE
	,EXIT_FAIL_TO_UNLINK_TMP_FILE
	,EXIT_ILLEGAL_CMD_ARGS =255
} EXIT_CODE;

/**
 * Declarations around "Parser for WEB Styled HTML files"
 **/
static int  cruncher(char * web_file_path); /* Parser Entry */

	/* Declare Tokens and Delimiters */
static char	 *BEGIN_TAG = "<" "pre"
		,*TAG_ATTR_FILEPATH = "file="
		,*END_TAG = "<" "/pre>"
		,EOL = '\0'
		,TAG_RPAREN = '>'
		,FILEPATH_QUOTE = '"';

	/* Define the Context of Parsing WEB Styled HTML Files */
typedef struct crunch_context_t {
	char	*web_file_path;	/* Path of the WEB Styled HTML File */
	FILE	*instream;	/* Input stream from "web_file_path" */
	char	line[LINE_MAX + 1]; /* Read line buffer */
	long int line_no;	/* Current read line number */
	FILE	*spit_out_stream; /* Output stream */
} CRUNCH_CONTEXT;

	/* Definition States of this Finite State Machine (FSM) */
typedef enum CRUNCH_STATES_t {
	 CRUNCH_IGNORE = 0
	,CRUNCH_SPIT_IT
	,MAX_CRUNCH_STATES
	,CRUNCH_ILLEGAL_STATE_CONTROL
} CRUNCH_STATES;

	/* The "CRUNCH_ACTION" is a pointer to a action-function. */
typedef CRUNCH_STATES (*CRUNCH_ACTION)(CRUNCH_CONTEXT *context,
					CRUNCH_STATES state);

	/* Declare Actions and FSM-table */
static CRUNCH_STATES seek_BEGIN_TAG(CRUNCH_CONTEXT *context,
					CRUNCH_STATES state);
static CRUNCH_STATES spit_it_out(CRUNCH_CONTEXT *context,
					CRUNCH_STATES state);
static CRUNCH_ACTION action_table[] = {
			 seek_BEGIN_TAG /* at CRUNCH_IGNORE */
			,spit_it_out	/* at CRUNCH_SPIT_IT */
		};

	/* CRUNCH_CONTEXT Dumper */
static void dump_FSM_context(FILE *stream, char *header,
				CRUNCH_CONTEXT *c);

	/* Declare methods depended on Tokens and Delimiters */
static int  strncmp_ignore_case(char * tag, char *line, size_t sz);
static char *search_spit_out_file_path(char *line);
static int  is_TAG_ATTR_FILEPATH_delimit(char code);

/**
 * Definition to update output files
 **/
#define	TMP_PATH	"/tmp/" /* Path to templates */
static pid_t	pid; /* For making template file names */
		/* Working buffer of template file path */
static size_t	tmp_path_size = (size_t)0L;
static char	*tmp_path = NULL;
static int	output_file_cnt = 0; /* Count of output file names */
static char	**output_files = NULL; /* List of output file names */

	/* Functions to update output files */
static char	*add_output_file_name(char *spit_out_file_path);
static char	*find_output_path(char *spit_out_file_path);
static char	*make_tmp_file_path(char *spit_out_file_path);
static EXIT_CODE update_outputs(void);
static void 	*map_file(char *path, int flags, struct stat *st);
static void	discharge_file_contents(char *source,
					void *sp, struct stat *st);
static void	create_output_file(char *dest, char *source,
					void *sp, struct stat *st);

/**
 * The Utilities
 **/
static void usage(char *command_name);
#	define	USAGE_FORMAT				\
	"CrunchWEB ($Revision: 1.10 $)\n"			\
	"Copyright 1999, Maoyam (maoyam@mail.goo.ne.jp)\n\n" \
	"This may be copied only under "		\
	"the GNU General Public License, "		\
	"which may be found in the CruncWEB source kit " \
	"(see "						\
	"http://sapporo.cool.ne.jp/maoya/crunchWEB/crunchWEB.html" \
	").\n\n" \
	"Usage: %s WEB_Styled_HTML_File_Path...\n"
static void print_err_WHICH_LINE(char *F, int L);
static void print_err(char * format, void *val);

/**
 * Program Entry...
 **/
int
main(int ac, char **av) {
	int	i, exit_code;

if (DEBUG>0) (void)fprintf(stderr,"+++++DEBUG=%d++: "
"Entry main(%d, 0x%x[0] = [%s])...\n", DEBUG, ac, av, av[0]);

	if (ac < 2) {
		usage(av[0]);
		exit_code = EXIT_ILLEGAL_CMD_ARGS;
	}
	else {
			/* `pid' for making template file names */
		pid	= getpid();

		for (exit_code = EXIT_NORMAL, i = 1; i < ac; ++i) {
			if (access(av[i], R_OK) != 0) {
				print_err_WHICH_LINE(__FILE__,__LINE__);
				print_err("Can't read av[%d] ",
							(void *)i);
				print_err("\"%s\"\n", (void *)av[i]);
			} else if ((exit_code = cruncher(av[i])) != 0) {
				break;
			}
		}

		if (exit_code == EXIT_NORMAL) {
			exit_code	= update_outputs();
		}
	}

if (DEBUG>0) (void)fprintf(stderr,"+++++DEBUG=%d++: "
"Exit from main(). Exit Code is %d.\n", DEBUG, exit_code);

	return	exit_code;
}

/**
 * Definitions for Parser of WEB Styled HTML files
 **/
static int
cruncher(char * web_file_path) {
	static CRUNCH_CONTEXT	context;
	CRUNCH_CONTEXT	*c = &context;
	CRUNCH_STATES	state, next_state;

		/* Initialize the context and FSM-state */
	state = CRUNCH_IGNORE;
	c->web_file_path	= web_file_path;
	c->spit_out_stream = (FILE *)NULL;

		/* Open the input stream */
	c->instream = fopen(web_file_path,"r");
	if (c->instream ==NULL) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Fail fopen(%s,\"r\")\n",
				(void *)c->web_file_path);
		return	(int)EXIT_FAIL_FOPEN;
	}

		/* Parser Loop */
	for (c->line_no = 1L;
		fgets((char *)memset((void *)c->line,
					(int)EOL,
					sizeof(c->line)),
			sizeof(c->line) - 1, c->instream) != NULL;
		++(c->line_no), state = next_state)
	{
		next_state = (action_table[state])(c, state);
		if (next_state == CRUNCH_ILLEGAL_STATE_CONTROL) {
			print_err_WHICH_LINE(__FILE__, __LINE__);
			print_err("Broken FSM-state Control "
				"(Last State = %ld)\n", (void *)state);
			dump_FSM_context(stderr, "\t", c);
			return	EXIT_BROKEN_FSM_STATE_CONTROL;
		}
	}

		/**
		 * Now, we found out *END_TAG*.
		 * We must close output stream has been opened file.
		 **/
	if ((c->spit_out_stream != stdout) && state != CRUNCH_IGNORE) {
		if (fclose(c->spit_out_stream) != 0) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Fail Close Output Stream (0x%x), "
				"at out of parser loop\n",
				(void *)c->spit_out_stream);
			perror("ERRNO MSG");

			return EXIT_FAIL_CLOSE_OUTPUT_OUT_OF_PARSE_LOOP;
		}
		c->spit_out_stream	= (FILE *)NULL;
	}

		/* Close the input stream */
	if (fclose(c->instream) != 0) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Fail Close Input Stream (0x%x), "
			"at out of parser loop\n",
			(void *)c->instream);
		perror("ERRNO MSG");

		return EXIT_FAIL_CLOSE_INPUT_STREAM;
	}
	c->instream	= (FILE *)NULL;

	return	EXIT_NORMAL;
}

static CRUNCH_STATES
seek_BEGIN_TAG(CRUNCH_CONTEXT *c, CRUNCH_STATES state) {
	if (state != CRUNCH_IGNORE) {
		state =	CRUNCH_ILLEGAL_STATE_CONTROL;
	}
	else if (strncmp_ignore_case(BEGIN_TAG, c->line,
			strlen(BEGIN_TAG)) == 0) {
		char * spit_out_file_path
			= search_spit_out_file_path(c->line);

		if (spit_out_file_path == NULL) {
			(void)fprintf(stdout,
				"----- %s:L%u -----\n",
				c->web_file_path, c->line_no);
			c->spit_out_stream	= stdout;
		}
		else {
			char	*tmp_file_path;

			tmp_file_path	=
				add_output_file_name(spit_out_file_path);
			if ((tmp_file_path == NULL)
			|| (c->spit_out_stream =
				fopen(tmp_file_path,
					(access(tmp_file_path, F_OK)
						== 0)? "a":"w")
			) == NULL) {
				print_err_WHICH_LINE(__FILE__,__LINE__);
				print_err("Can't create/wite-open "
						"tmp-file \"%s\"\n",
					(void *)tmp_file_path);

				(void)fprintf(stdout,
					"----- In %s: Line# %u "
					"(Tmp-file Access Denied): "
					"Output File is \"%s\" -----\n",
						c->web_file_path,
						c->line_no,
						spit_out_file_path);
				c->spit_out_stream	= stdout;
			}
		}

		state	= CRUNCH_SPIT_IT;
	}
	else {
		/* Ignore this read line buffer */
		state = CRUNCH_IGNORE;
	}
	
	return	state;
}

static CRUNCH_STATES
spit_it_out(CRUNCH_CONTEXT *c, CRUNCH_STATES state) {
	if (state != CRUNCH_SPIT_IT) {
		state	= CRUNCH_ILLEGAL_STATE_CONTROL;
	}
	else if (strncmp_ignore_case(END_TAG, c->line,
			strlen(END_TAG)) == 0) {
		if ((c->spit_out_stream != stdout)
		&& fclose(c->spit_out_stream) != 0) {
			print_err_WHICH_LINE(__FILE__,
					__LINE__);
			print_err("Fail Close Output "
				"Stream (0x%x)\n",
				(void *)c->spit_out_stream);
			perror("ERRNO MSG");

			return	EXIT_FAIL_CLOSE_OUTPUT;
		}
		c->spit_out_stream	= (FILE *)NULL;

		state	= CRUNCH_IGNORE;
	}
	else {
			/* Write the read buffer to output stream */
		if (fputs(c->line, c->spit_out_stream) == EOF)
		{
			print_err_WHICH_LINE(__FILE__,
					__LINE__);
			print_err("Fail Write Out "
				"[%s] ", (void *)c->line);
			print_err("into the output "
				"stream (0x%x)\n",
				(void *)c->spit_out_stream);
			perror("ERRNO MSG");

			return	EXIT_FAIL_FPUTS;
		}

		state	= CRUNCH_SPIT_IT;
	}

	return	state;
}

static int
strncmp_ignore_case(char *tag, char *line, size_t sz) {
	int	result;
	size_t	i;

	for (i = 0; i < sz && line[i] != EOL; ++i) {
		if (tag[i] == EOL) {
			result	= -1; /* tag[] < line[] */
			break;
		}

		result = toupper((int)tag[i]) - toupper((int)line[i]);
		if (result != 0) break;
	}

	return	result;
}

static char *
search_spit_out_file_path(char *line) {
	char	*result;

	for (line += strlen(BEGIN_TAG);
		*line != EOL && isspace((int)*line); ++line);
	if (*line == EOL
	|| strncmp_ignore_case(TAG_ATTR_FILEPATH, line,
				strlen(TAG_ATTR_FILEPATH)) != 0) {
		/* Unfound/Exist Illegal PRE-Tag Attribute */
		return	NULL;
	}

	line	+= strlen(TAG_ATTR_FILEPATH);
	if (*line == FILEPATH_QUOTE) {
		++line;
	}
	for (result = line;
		!is_TAG_ATTR_FILEPATH_delimit(*line); ++line) ;
	*line	= EOL;

	return	result;
}

static int
is_TAG_ATTR_FILEPATH_delimit(char code) {
	return	code == EOL
		|| code == TAG_RPAREN
		|| code == FILEPATH_QUOTE
		|| isspace((int)code);
}

static void
dump_FSM_context(FILE *stream, char *header, CRUNCH_CONTEXT *c) {
	fprintf(stream,"%s<<< Contents of CRUNCH_CONTEXT >>>\n",
			header);
	fprintf(stream,"%s   web_file_path   = [%s]\n",
			header, c->web_file_path);
	fprintf(stream,"%s   instream        = 0x%x\n",
			header, c->instream);
	fprintf(stream,"%s   line[]          = [%s]\n",
			header, c->line);
	fprintf(stream,"%s   line_no         = %ld\n",
			header, c->line_no);
	fprintf(stream,"%s   spit_out_stream = 0x%x (stdout = 0x%x)\n",
			header, c->spit_out_stream);
}
static char *
add_output_file_name(char *spit_out_file_path) {
	if ((output_files == NULL)
	|| find_output_path(spit_out_file_path) == NULL) {
		if ((output_files = (char **)realloc(
			(void *)output_files,
			sizeof(output_files) * ++output_file_cnt)
		) == (char **)NULL) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Can't allocate `output_files' "
				"(output_file_cnt = %d)\n",
					(void *)output_file_cnt);

			return	NULL;
		}

		if ((output_files[output_file_cnt - 1]	=
			(char *)malloc(strlen(spit_out_file_path) + 1)
		) == NULL) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Can't allocate areas of "
				"output file path (%s).\n",
				(void *)spit_out_file_path);

			return	NULL;
		}

		strcpy(output_files[output_file_cnt - 1],
						spit_out_file_path);
	}

	return	make_tmp_file_path(spit_out_file_path);
}

static char	*find_output_path(char *spit_out_file_path) {
	char	**found;
	int	i;

	for (found = output_files, i = 0; i < output_file_cnt;
							++i, ++found) {
		if (strcmp(*found, spit_out_file_path) == 0) {
			return	*found; /* The file is stored, yet. */
		}
	}

	return	NULL; /* The file is new one. */
}

static char	*
make_tmp_file_path(char *spit_out_file_path) {
	size_t	spit_out_file_path_sz =
			(size_t)(strlen(TMP_PATH)
			+ strlen(spit_out_file_path)
			+ (sizeof(pid) * 2) + 2);

	if (tmp_path_size < spit_out_file_path_sz) {
		if ((tmp_path = (char *)realloc(tmp_path,
					spit_out_file_path_sz))
		== NULL) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Can't allocate `tmp_path'. "
				"`pid' = 0x%x", (void *)pid);
			print_err("(spit_out_file_path = \"%s\")\n",
				(void *)spit_out_file_path);
		}

		tmp_path_size	= spit_out_file_path_sz;
	}

	(void)sprintf(tmp_path, TMP_PATH "%s.%x",
					spit_out_file_path, pid);

	return	tmp_path;
}

static EXIT_CODE
update_outputs(void) {
	int	i, s_desc;
	char	**dest, *source;
	struct stat	s_st;
	void	*sp;

	for (i = 0, dest = output_files; i < output_file_cnt;
							++i, ++dest) {
			/* Get the tmp-file path of `dest' file */
		source	= make_tmp_file_path(*dest);

			/* Mapping the `source' file to memory */
		if ((sp = map_file(source, O_RDWR, &s_st)) == NULL) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Fail to mapping `source (%s)'.\n",
				(void *)source);

			return	EXIT_FAIL_TO_MMAP_TMP_FILE;
		}

		if ((access(*dest, F_OK) == 0)
		&& access(*dest, W_OK) != 0) {
				/**
				 * If `dest' file is write-protected,
				 * Discharge the tmp-file into `stdout'
				 **/
			discharge_file_contents(source, sp, &s_st);
		}
		else {
				/* Try to write out `dest' file */
			create_output_file(*dest, source, sp, &s_st);
		}

			/* Unmapping `source' file */
		if (munmap(sp, s_st.st_size) != 0) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Fail to unmapping `source (%s)' ",
							(void *)source);
			print_err("mapping to 0x%x.\n", (void *)sp);
			perror("ERRNO MSG");

			return	EXIT_FAIL_TO_UNMAP_TMP_FILE;
		}

		if (unlink(source) != 0) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Fail to unlink `source (%s)'.\n",
							(void *)source);
			perror("ERRNO MSG");

			return	EXIT_FAIL_TO_UNLINK_TMP_FILE;
		}
	}

	return	EXIT_NORMAL;
}

static void *
map_file(char *path, int flags, struct stat *st) {
	int	desc;
	void	*p = NULL;

	if ((desc = open(path, O_RDWR | flags)) == -1) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Can't open (flags = 0x%x) ",
			(void *)(O_RDWR | flags));
		print_err("\"%s\" file.\n", (void *)path);
		perror("ERRNO MSG");
	}
	else if (fstat(desc, st) != 0) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Can't get stat-struct (at 0x%x) ",
							(void *)st);
		print_err("of \"%s ", (void *)path);
		print_err("(descriptor = %d)\" file.\n", (void *)desc);
		perror("ERRNO MSG");
	}
	else if ((p = mmap(0, st->st_size,
			(PROT_READ|PROT_WRITE), MAP_PRIVATE, desc, 0)
	) == MAP_FAILED) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Can't memory-mapping \"%s ", (void *)path);
		print_err("(descriptor = %d)\" file. ", (void *)desc);
		print_err("Size of the file is %u.\n",
						(void *)(st->st_size));
		perror("ERRNO MSG");

		p	= NULL; /* `MAP_FAILED' is not `NULL'. */
	}

	if (desc != -1)	close(desc);

	return	p;
}

static void	discharge_file_contents(char *source,
					void *sp, struct stat *st) {
	printf("--- %s ---\n", source);

	if (write(STDOUT_FILENO, sp, st->st_size) < st->st_size) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Fail to discharge `source (%s, ",
							(void *)source);
		print_err("mapping at 0x%x)' into `stdout'.  ",
							(void *)sp);
		print_err("The `source' size is %u.\n",
						(void *)(st->st_size));
		perror("ERRNO MSG");
	}

	return;
}

static void
create_output_file(char *dest, char *source,
					void *sp, struct stat *st) {
	int		desc;
	struct stat	d_st;
	void		*dp;

		/* When `dest' file is not exist, We must reate it! */
	d_st.st_size	= 0; dp	= NULL;

		/* Memory-mapping `dest' file */
	if ((((access(dest,F_OK) == 0)	/* `dest' is exist */
	&& access(dest, W_OK) == 0)	/* `dest' is writable */
	&& access(dest, R_OK) == 0)	/* `dest' is readable */
	&& (dp = map_file(dest, O_RDWR, &d_st)) == NULL) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Fail to mapping `dest (%s).\n",
						(void *)dest);
		perror("ERRNO MSG");

		discharge_file_contents(source, sp, st);
	}
	else if (d_st.st_size != st->st_size) {
			/**
			 * If sizes of `dest' and `source' are differ,
			 * Create the `dest' file.
			 **/
		if ((desc = open(dest, O_CREAT|O_WRONLY|O_TRUNC,
				(S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP))
		) == -1) {
			print_err_WHICH_LINE(__FILE__,__LINE__);
			print_err("Fail to open `dest (%s)' "
				"for write only, create & "
				"truncate.\n", (void *)dest);
			perror("ERRNO MSG");

			discharge_file_contents(source, sp, st);
		}
		else {
			if (write(desc, sp, st->st_size)
			!= st->st_size) {
				print_err_WHICH_LINE(__FILE__,__LINE__);
				print_err("Fail to write out "
					"`dest (%s)' file.  ", (void *)dest);
				print_err("The size of `source (%s, ",
							(void *)source);
				print_err("mapping at 0x%x)' ",
							(void *)sp);
				print_err("is %u.\n",
						(void *)(st->st_size));
				perror("ERRNO MSG");

				discharge_file_contents(source, sp, st);
			}

			if (close(desc) == -1) {
				print_err_WHICH_LINE(__FILE__,__LINE__);
				print_err("Fail to close "
					"`dest (%s)'.\n", (void *)dest);
				perror("ERRNO MSG");
			}
		}
	}
	else if (memcmp(dp, sp, st->st_size) != 0) {
			/**
			 * If the mapping areas of `dest' and `source'
			 * are differ,  Memory-copy between those ares.
			 **/
		(void)memcpy(dp, sp, st->st_size);
	}

	if ((dp != NULL) && munmap(dp, d_st.st_size) != 0) {
		print_err_WHICH_LINE(__FILE__,__LINE__);
		print_err("Fail to unmapping `dest (%s, ",
							(void *)dest);
		print_err("mapping at 0x%x)'.\n", (void *)dp);
		perror("ERRNO MSG");
	}

	return;
}

/**
 * Definitions of the Utilities
 **/
static void
usage(char *command_name) {
	print_err(USAGE_FORMAT, (void *)command_name);
}

static void
print_err_WHICH_LINE(char *F, int L) {
	(void)fprintf(stderr,"Oops! %s:",F);
	(void)fprintf(stderr, "%d: ",L);
}

static void
print_err(char * format, void *val) {
	(void)fprintf(stderr, format, val);
}

