// Kauffman.java (Java applet)
//
//	  This applet's idea comes from a book "AT HOME IN THE
//	UNIVERSE" wriiten by Stuart Kauffman (the 4th section).
//
// (C) by M.Yamada since Sun Nov 14 13:35:55 JST 1999
//
// Change Logs:
//   Ver1.00 Original by M.Yamada in Sun Nov 14 13:36:20 JST 1999
//   Vwr1.01 Parameter-K can be set between 1 and 3.
//			by M.Yamada in Tue Nov 23 19:12:57 JST 1999

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Random;

public class Kauffman extends Applet implements Runnable {
	static final private long INTERVAL_MILLI_SEC = 200L;
	Thread	t;
	boolean	paint_done;

	Dimension	sz;
	Image		bs;
	Graphics	bs_g;
	Color		on_color = Color.blue;

	static final private int LATTICE = 100;
	static final private double LATTICE_D = 100.0;
	private int PIXCEL_WIDTH = 1;
	Random	random;
	int	start_cnt;
	int	K_param; // K-parameter (1 <= K <= 3; K is integer)
	int	Step_Forward;	// Refreshing Display Step Count
	int	max_funcs; // K = {1, 2, 3} => max_funcs = {2, 14, 254}
	Bool_Func	funcs[] = new Bool_Func[254];
					// 2^(2^3) - 2 = 256 - 2 = 254
	Node	nodes[][];
	int	ages[][];
	byte	nexts[][];

	Color_table color_table;

	class Bool_Func {
		private byte results[][][] = new byte[2][2][2];

			// Constructor
		public Bool_Func(byte r[]) {
			int i, j, k, maxi, maxj, maxk = 2;

			maxi	= (r.length > 4)? 2: 1;
			maxj	= (r.length < 4)? 1: 2;

			for (i = 0; i < maxi; ++i)
			for (j = 0; j < maxj; ++j)
			for (k = 0; k < maxk; ++k)
				results[i][j][k] = r[i * 4 + j * 2 + k];
		}

			// Returns result of a Bool-Functions (2 inputs)
		public byte eval(byte i, byte j, byte k) {
			return	results[i][j][k];
		}
	}

	class Node {
		private Node	input[] = new Node[3];
		private byte		on_off;
		private Bool_Func	func;

			// Constructors
		public Node(byte init_state) {
			on_off	= init_state;
			input[0] = input[1]	= null;
			func	= null;
		}
		public Node() {
			this((byte)0);
		}

			// Utilities
		public void set_inputs(Node n0, Node n1, Node n2) {
			input[0]	= n0;
			input[1]	= n1;
			input[2]	= n2;
		}

		public void set_state(byte s) { on_off = s; }
		public byte get_state() {
			return on_off;
		}

		public void set_Bool_Func(Bool_Func f) {
			func	= f;
		}

		public byte next_state() {
			return	func.eval(
				(input[0] == null)?
					(byte)0: input[0].get_state(),
				(input[1] == null)?
					(byte)0: input[1].get_state(),
				input[2].get_state());
		}
	}

	class Color_table {
		private Color colors[] = new Color[256];

		public Color_table() {
			colors[0]	= Color.black;
			for (int i = 1; i < 256; ++i) {
				if (i < 86) {
					colors[i] = new Color(
						i + 99,	// Red
						0,	// Green
						256 - i); // Blue
				}
				else if (i < 171) {
					colors[i] = new Color(
						341 - i,
						i + 14,
						0);
				}
				else {
					colors[i] = new Color(
						0,
						426 - i,
						i - 71);
				}
			}
		}

		public Color int_to_color(int a) {
			return colors[(a == 0)?
				0:
				((a % 255) + ((a < 255)? 0: 1))];
		}
	}

	public void reset_nodes() {
		int	i, j;
		for (i = 0; i < LATTICE; ++i)
		for (j = 0; j < LATTICE; ++j) {
			ages[i][j]	=
				(new Double(Math.floor(
					random.nextDouble() * 2.0D))
				).intValue();
			nodes[i][j].set_Bool_Func(funcs[
				(new Double(Math.floor(
					random.nextDouble() *
						(new Integer(max_funcs))
							.doubleValue()
				))).intValue()]);
		}

		int	u, v, w, x, y, z;
		for (i = 0; i < LATTICE; ++i)
		for (j = 0; j < LATTICE; ++j) {
			u = (new Double(Math.floor(
					random.nextDouble() *
						LATTICE_D))).intValue();
			v = (new Double(Math.floor(
					random.nextDouble() *
						LATTICE_D))).intValue();
			w = (new Double(Math.floor(
					random.nextDouble() *
						LATTICE_D))).intValue();
			x = (new Double(Math.floor(
					random.nextDouble() *
						LATTICE_D))).intValue();
			y = (new Double(Math.floor(
					random.nextDouble() *
						LATTICE_D))).intValue();
			z = (new Double(Math.floor(
					random.nextDouble() *
						LATTICE_D))).intValue();

			nodes[i][j].set_inputs(
				(K_param < 3)? null: nodes[u][v],
				(K_param < 2)? null: nodes[w][x],
				nodes[y][z]);
		}
	}

	public void init() {
		random	= new Random(System.currentTimeMillis());
		sz	= this.getSize();
		if (sz.width > LATTICE && sz.height > LATTICE)
			PIXCEL_WIDTH	=
				((sz.width < sz.height)?
					sz.width: sz.height) / 100;

		sz	= new Dimension(LATTICE * PIXCEL_WIDTH,
					LATTICE * PIXCEL_WIDTH);
		bs	= createImage(sz.width, sz.height);
		bs_g	= bs.getGraphics();

		String	K = this.getParameter("K-Parameter");
		try { K_param = Integer.parseInt(K); }
		catch (Exception e) { K_param	= 2; }
		K	= this.getParameter("Step_Forward");
		try { Step_Forward = Integer.parseInt(K); }
		catch (Exception e) { Step_Forward	= 7; }

		byte	results[];
		switch (K_param) {
		case 1:
			results = new byte[2];
			max_funcs	= 2;
			break;
		case 2:
			results	= new byte[4];
			max_funcs	= 14;
			break;
		case 3:
		default:
			results	= new byte[8];
			max_funcs	= 254;
			break;
		}

		int	i, j;
		for (i = 0; i < max_funcs; ++i) {
			for (j = 0; j < results.length; ++j)
				results[j]
					= (byte)(((i + 1) >> j) & 0x01);

			funcs[i] = new Bool_Func(results);
		}

		nodes = new Node[LATTICE][LATTICE];
		ages = new int[LATTICE][LATTICE];
		nexts = new byte[LATTICE][LATTICE];

		for (i = 0; i < LATTICE; ++i)
		for (j = 0; j < LATTICE; ++j) {
			nodes[i][j] = new Node();
		}

		reset_nodes();	start_cnt = 0;

		color_table	= new Color_table();
	}

	public void update(Graphics g) { paint(g); }
	public void paint(Graphics g) {
		int	i, j, k;
		for (k = 0; k < Step_Forward; ++k) {
			for (i = 0; i < LATTICE; ++i)
			for (j = 0; j < LATTICE; ++j) {
				nexts[i][j] = nodes[i][j].next_state();
				if ((ages[i][j] != Integer.MAX_VALUE)
				&& nexts[i][j] != 0)
					++ages[i][j];
				else
					ages[i][j]	= 0;
			}

			for (i = 0; i < LATTICE; ++i)
			for (j = 0; j < LATTICE; ++j) {
				nodes[i][j].set_state(nexts[i][j]);
			}

			if (t == null)	break;
		}

		for (i = 0; i < LATTICE; ++i)
		for (j = 0; j < LATTICE; ++j) {
			bs_g.setColor(color_table.int_to_color(
							ages[i][j]));
			bs_g.fillRect(i * PIXCEL_WIDTH,
					j * PIXCEL_WIDTH,
					PIXCEL_WIDTH, PIXCEL_WIDTH);
		}

		g.drawImage(bs, 0, 0, this);
		paint_done	= true;
	}

	public void run() {
		while (t != null) {
			paint_done	= false;
			repaint();
			while (!paint_done) {
				try {
					Thread.sleep(300L);
				}
				catch (InterruptedException e) { ; }
				if (t == null)	return;
			}

			try { Thread.sleep(INTERVAL_MILLI_SEC); }
			catch (InterruptedException e) { ; }
		}
	}

	public void start() {
		if (t == null) {
			if (start_cnt++ > 0) 
				reset_nodes();

			t	= new Thread(this);
			t.start();
		}
	}

	public void stop() {
		if (t != null) {
			t.stop();
			t = null;
		}
	}
}
